//   sha256
// (c) Proteus
// icq: 133575351  email:lawnmower-man@mail.ru 
//
//          Intel-  
//         
// ,       // #

#include<stdio.h>
#include<memory.h>
#include<stdlib.h>
#include<assert.h>
#include<math.h>

typedef unsigned char byte;
typedef unsigned long dword;
typedef unsigned __int64 dword64;
typedef unsigned int uint;

namespace _SHA2{
	dword k[64];
/******
static const dword k[64]=
   ( 32       64-   [ 2  311]):
   0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5, 0x3956C25B, 0x59F111F1, 0x923F82A4, 0xAB1C5ED5,
   0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3, 0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174,
   0xE49B69C1, 0xEFBE4786, 0x0FC19DC6, 0x240CA1CC, 0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA,
   0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7, 0xC6E00BF3, 0xD5A79147, 0x06CA6351, 0x14292967,
   0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13, 0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85,
   0xA2BFE8A1, 0xA81A664B, 0xC24B8B70, 0xC76C51A3, 0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070,
   0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5, 0x391C0CB3, 0x4ED8AA4A, 0x5B9CCA4F, 0x682E6FF3,
   0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208, 0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2

static const dword state_ini[8]=
   ( 32          [ 2  19])
   0x6A09E667, 0xBB67AE85, 0x3C6EF372, 0xA54FF53A, 0x510E527F, 0x9B05688C, 0x1F83D9AB, 0x5BE0CD19

***/
}


//! hash: sha
class SHA256{
	union{
		dword state[8];		//!<     (a,b,c,d,e,f,g,h)
		byte output[32];	//!<   
	};
	union{
		byte buffer[64];	//!<    
		dword  data[64];	//!<    
	};
	uint size;				//!<  
	uint total_size;		//!<    
	void add(byte);			//!<    

	void init();
	void final(byte[32]);
	void hash_block();		//!<   
	static void install(dword*,uint,dword,int);	//!  
public:
	static void install();	//!<   (  ,    )
	void hash(const char*,byte res[32]);
};

//!  
//!      ,       
//!                
//! \retval res   
//! \param count   
//! \param check   
//! \param deg   (2,3)
void SHA256::install(dword *res,uint count,dword check,int deg)
{
	assert(count<=64);
	uint n;
	dword p,s,d[64]; //    
	d[0]=2; p=1;
	for(n=1; n<count; n++) { //     
		p+=2;
		for(uint m=0; m<n;) if(!(p%d[m])) p+=2,m=0; else m++;
		d[n]=p;
	}
	//    
	for(n=0; n<count; n++) res[n]=(dword)ldexp(fmod(pow(d[n],1.0/deg),1),32);
	for(n=0,s=0; n<count; n++) s+=res[n];
	assert(s==check);
}


void SHA256::install()
{
	install(_SHA2::k,64,0x941d1755,3);
}


void SHA256::init()
{
	size=total_size=0;
	install(state,8,0x6ea8df67,2);
}


void SHA256::add(byte x)
{
	total_size++;
	buffer[size++^3]=x; //#
	if(size>=64) hash_block(),size=0;
}


void SHA256::final(byte res[32]) 
{
	uint n;
	dword64 total=dword64(total_size)*8;
	//   
	add(0x80); while(size!=56) add(0); 
	//       
	const byte *t=(byte*)&total;
	for(n=8; n;) add(t[--n]);
	//   ,  
	for(n=0; n<32; n++) res[n]=output[n^3];	//#
}


void SHA256::hash_block() 
{
	uint n;
	dword a=state[0];
	dword b=state[1];
	dword c=state[2];
	dword d=state[3];
	dword e=state[4];
	dword f=state[5];
	dword g=state[6];
	dword h=state[7];
	//   ()
	for(n=16; n<64; n++) {
		dword s1=data[n-15]; s1=_rotr(s1, 7)^_rotr(s1,18)^(s1>> 3);
		dword s2=data[n- 2]; s2=_rotr(s2,17)^_rotr(s2,19)^(s2>>10);
		data[n]=data[n-16]+data[n-7]+s1+s2;
	}
	//  
	for(n=0; n<64; n++) {
		dword aa=_rotr(a,2)^_rotr(a,13)^_rotr(a,22);
		dword ee=_rotr(e,6)^_rotr(e,11)^_rotr(e,25);
		dword t1=h+ee+ _SHA2::k[n]+data[n]+(g^e&(f^g)); // ((e&f)^(~e&g));
		dword t2=aa+((a&b)^(a&c)^(b&c));
		h=g, g=f, f=e, e=d+t1, d=c, c=b, b=a, a=t1+t2;
	}
	state[0]+=a;
	state[1]+=b;
	state[2]+=c;
	state[3]+=d;
	state[4]+=e;
	state[5]+=f;
	state[6]+=g;
	state[7]+=h;
}


void SHA256::hash(const char* str,byte res[32])
{
	init();
	while(*str) add(*str++);
	final(res);
}



void main()
{
	assert(sizeof(byte)==1 && sizeof(dword)==4 && sizeof(dword64)==8);

	byte res[32];
	byte test[32]={
		0x65,0xe8,0x4b,0xe3,0x35,0x32,0xfb,0x78,0x4c,0x48,0x12,0x96,0x75,0xf9,0xef,0xf3,
		0xa6,0x82,0xb2,0x71,0x68,0xc0,0xea,0x74,0x4b,0x2c,0xf5,0x8e,0xe0,0x23,0x37,0xc5 
	};
	SHA256::install();
	SHA256().hash("qwerty",res);
	if(memcmp(res,test,32)) printf("Internal error!\n\a"),getchar();
	printf("done..\n");
}

